"""Formula 8.15 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form8Dot15PrestressTransferStress(Formula):
    """Class representing formula 8.15 for the calculation of the constant bond stress at which prestress is assumed to be transferred to the
    concrete, at the release of tendons.
    """

    label = "8.15"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        eta_p1: DIMENSIONLESS,
        eta_1: DIMENSIONLESS,
        f_ctd_t: MPA,
    ) -> None:
        """[:math:`f_{bpt}`] Constant bond stress at which prestress is assumed to be transferred to the concrete [:math:`MPa`].

        NEN-EN 1992-1-1+C2:2011 art.8.10.2.2(1) - Formula (8.15)

        Parameters
        ----------
        eta_p1 : DIMENSIONLESS
            [:math:`η_{p1}`] Coefficient that takes into account the type of tendon and the bond situation at release [-].

            = 2.7 for indented wires.

            = 3.2 for 3 and 7-wire strands

            Use your own implementation for this value or use :class:`SubForm8Dot15EtaP1` class.
        eta_1 : DIMENSIONLESS
            [:math:`η_1`] Coefficient related to the quality of the bond condition and the position of the bar during concreting (see Figure 8.2) [-].

            = 1 when ‘good’ conditions are obtained;

            = 0.7 other cases and for bars in structural elements built with slip-forms, unless it can be shown that ‘good’ bond conditions exist;

            Use your own implementation of this formula or use the :class:`SubForm8Dot2CoefficientQualityOfBond` class.
        f_ctd_t : MPA
            [:math:`f_{ctd}(t)`] Design tensile value of strength at time of release [:math:`mm`].

            = :math:`α_{ct} ⋅ 0.7 ⋅ f_{ctm}(t) / γ_{c}` (see 3.1.2(9) and 3.1.6(2)P)

            Use your own implementation for this value or use :class:`SubForm8Dot15TensileStrengthAtRelease` class.
        """
        super().__init__()
        self.eta_p1 = eta_p1
        self.eta_1 = eta_1
        self.f_ctd_t = f_ctd_t

    @staticmethod
    def _evaluate(
        eta_p1: DIMENSIONLESS,
        eta_1: DIMENSIONLESS,
        f_ctd_t: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            eta_p1=eta_p1,
            eta_1=eta_1,
            f_ctd_t=f_ctd_t,
        )
        return eta_p1 * eta_1 * f_ctd_t


class SubForm8Dot15EtaP1(Formula):
    """Class representing sub-formula 8.15 for the calculation of the coefficient that takes into account the type of tendon and the bond situation
    at release.
    """

    label = "8.15"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        type_of_wire: str,
    ) -> None:
        """[:math:`η_{p1}`] Coefficient that takes into account the type of tendon and the bond situation at release [-].

        NEN-EN 1992-1-1+C2:2011 art.8.10.2.2(1) - Formula (8.15)

        Parameters
        ----------
        type_of_wire : str
            Type of wire.

            = 'indented' for indented wires;

            = '3_7_wire_strands' for 3 and 7-wire strands;
        """
        super().__init__()
        self.type_of_wire = type_of_wire

    @staticmethod
    def _evaluate(type_of_wire: str) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        match type_of_wire.lower():
            case "indented":
                return 2.7
            case "3_7_wire_strands":
                return 3.2
            case _:
                raise ValueError(f"Invalid type of wire: {type_of_wire}. Options: 'indented' or '3_7_wire_strands'")


class SubForm8Dot15TensileStrengthAtRelease(Formula):
    """Class representing sub-formula 8.15 for the calculation of the design tensile value of strength at time of release (see  3.1.2(8) and
    3.1.6(2)P).
    """

    label = "8.15"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        alpha_ct: DIMENSIONLESS,
        f_ctm_t: MPA,
        gamma_c: DIMENSIONLESS,
    ) -> None:
        """[:math:`f_{ctd}(t)`] Design tensile value of strength at time of release [:math:`MPa`].

        NEN-EN 1992-1-1+C2:2011 art.8.10.2.2(1) - Formula (8.15)

        Parameters
        ----------
        alpha_ct : DIMENSIONLESS
            [:math:`α_{ct}`] coefficient taking account of long term effects on the tensile strength and of unfavourable effects, resulting from the
            way the load is applied. [-].

            Value may be found in national annex. Recommended value: 1.0
        f_ctm_t : MPA
            [:math:`f_{ctm}(t)`] Mean value of tensile strength at time of release (see formula 3.4) [:math:`MPa`].

            Use your own implementation for this value or use :class:`Form3Dot4DevelopmentTensileStrength` class.
        gamma_c : DIMENSIONLESS
            [:math:`γ_{c}`] Partial safety factor for concrete [-].
        """
        super().__init__()
        self.alpha_ct = alpha_ct
        self.f_ctm_t = f_ctm_t
        self.gamma_c = gamma_c

    @staticmethod
    def _evaluate(
        alpha_ct: DIMENSIONLESS,
        f_ctm_t: MPA,
        gamma_c: DIMENSIONLESS,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(alpha_ct=alpha_ct, f_ctm_t=f_ctm_t)
        raise_if_less_or_equal_to_zero(gamma_c=gamma_c)
        return alpha_ct * 0.7 * f_ctm_t / gamma_c



"""Formula 3.2 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import DAYS


class Form3Dot2CoefficientDependentOfConcreteAge(Formula):
    """Class representing formula 3.2 for the coefficient βcc(t) which is dependent of the age of concrete."""

    label = "3.2"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        s: float,
        t: DAYS,
    ) -> None:
        """Calculates beta_cc(t) coefficient which is dependent of the age of concrete in days [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.2(6) - Formula (3.2)

        Parameters
        ----------
        s : float
            [s] Coefficient dependent on the kind of cement [-].
            = 0.20 for cement of strength classes CEM 42.5 R, CEM 52.5 N, and CEM 52.5 R (class R);
            = 0.25 for cement of strength classes CEM 32.5 R, CEM 42.5 N (class N);
            = 0.38 for cement of strength class CEM 32.5 N (class S).
            Use your own implementation of this formula or use the SubForm3Dot2CoefficientTypeOfCementS class.
        t : DAYS
            [t] Age of concrete in days [days].

        Returns
        -------
        None
        """
        super().__init__()
        self.s = s
        self.t = t

    @staticmethod
    def _evaluate(
        s: float,
        t: DAYS,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if s < 0:
            raise ValueError(f"Invalid s: {s}. s cannot be negative")
        if t <= 0:
            raise ValueError(f"Invalid t: {t}. t cannot be negative or zero")
        return np.exp(s * (1 - (28 / t) ** (1 / 2)))


class SubForm3Dot2CoefficientTypeOfCementS(Formula):
    """Class representing sub-formula for formula 3.2, which calculates the coefficient 's' which is dependent on the cement class."""

    source_document = NEN_EN_1992_1_1_C2_2011
    label = "3.2"

    def __init__(
        self,
        cement_class: str,
    ) -> None:
        """[s] Coefficient that depends on the type of cement [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.2(6) - s

        Parameters
        ----------
        cement_class : str
            [cement_class] Class of the cement.
                = 'R' for cement of strength classes CEM 42.5 R, CEM 52.5 N, and CEM 52.5 R (class R);
                = 'N' for cement of strength classes CEM 32.5 R, CEM 42.5 N (class N);
                = 'S' for cement of strength class CEM 32.5 N (class S).

        """
        super().__init__()
        self.cement_class = cement_class

    @staticmethod
    def _evaluate(
        cement_class: str,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        match cement_class.lower():
            case "r":
                return 0.20
            case "n":
                return 0.25
            case "s":
                return 0.38
            case _:
                raise ValueError(f"Invalid cement class: {cement_class}. Options: 'R', 'N' or 'S'")



"""Sub-formula 3.28, 3.29 and 3.30 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA


class SubForm3Dot282930Mu(Formula):
    """Class representing sub-formula for 3.28, 3.29 and 3.30 for the calculation of μ."""

    label = "3.28 - 3.29 - 3.30"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        sigma_pi: MPA,
        f_pk: MPA,
    ) -> None:
        """[μ] Ratio between initial pre-stress and characteristic tensile strength  (σpi / fpk) [-].

        NEN-EN 1992-1-1+C2:2011 art.3.3.2(7) - μ

        Parameters
        ----------
        sigma_pi : MPA
            [σpi] Initial pre-stress [MPa]
        f_pk : MPA
            [fpk] Characteristic tensile strength of pre-stress steel [MPa]

        Returns
        -------
        None
        """
        super().__init__()
        self.sigma_pi = sigma_pi
        self.f_pk = f_pk

    @staticmethod
    def _evaluate(
        sigma_pi: MPA,
        f_pk: MPA,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_pk < 0:
            raise ValueError(f"Invalid f_pk: {f_pk}. f_pk cannot be negative")
        return sigma_pi / f_pk



"""Module for unit conversions inside of Blueprints."""

# <editor-fold desc="Forces conversion">
KN_TO_N = 1e3
N_TO_KN = 1e-3
MM3_TO_M3 = 1e-9
MPA_TO_KPA = 1e3
# </editor-fold>

# <editor-fold desc="Section modulus conversion">
CM3_TO_MM3 = 1e3
# </editor-fold>

# <editor-fold desc="Moments conversion">
NMM_TO_KNM = 1e-6
# </editor-fold>



"""Formula 5.8 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import M
from blueprints.validations import raise_if_negative


class Form5Dot8EffectiveSpan(Formula):
    """Class representing formula 5.8 for calculating the effective span of beams and slabs, :math:`l_{eff}`.

    See Figure 5.4
    """

    label = "5.8"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        l_n: M,
        a_1: M,
        a_2: M,
    ) -> None:
        """[:math:`l_{eff}`] the effective span of a member [:math:`m`].

        NEN-EN 1992-1-1+C2:2011 art.5.3.2.2(1) - Formula (5.8)

        Parameters
        ----------
        l_n : M
            [:math:`l_{n}`] clear distance between the faces of the supports [:math:`m`].
        a_1 : M
            [:math:`a_{1}`] values for :math:`a_{1}` and :math:`a_{2}` at each end of the span, may be determined from the appropriate :math:`a_{i}`
                            values in Figure 5.4 where t is the width of the supporting element as shown. [:math:`m`].
        a_2 : M
            [:math:`a_{2}`] values for :math:`a_{1}` and :math:`a_{2}` at each end of the span, may be determined from the appropriate :math:`a_{i}`
                            values in Figure 5.4 where t is the width of the supporting element as shown. [:math:`m`].
        """
        super().__init__()
        self.l_n = l_n
        self.a_1 = a_1
        self.a_2 = a_2

    @staticmethod
    def _evaluate(
        l_n: M,
        a_1: M,
        a_2: M,
    ) -> M:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            l_n=l_n,
            a_1=a_1,
            a_2=a_2,
        )
        return l_n + a_1 + a_2

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.8."""
        return LatexFormula(
            return_symbol=r"l_{eff}",
            result=f"{self:.3f}",
            equation=r"l_{n} + a_{1} + a_{2}",
            numeric_equation=f"{self.l_n:.3f} + {self.a_1:.3f} + {self.a_2:.3f}",
            comparison_operator_label="=",
        )



"""Formula 5.1 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot2Eccentricity(Formula):
    """Class representing formula 5.2 for the calculation of eccentricity, ei."""

    label = "5.2"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        theta_i: DIMENSIONLESS,
        l_0: M,
    ) -> None:
        """[ei] Eccentricity, ei, for isolated members [m].

        NEN-EN 1992-1-1+C2:2011 art.5.2(7) - Formula (5.2)

        Parameters
        ----------
        theta_i : DIMENSIONLESS
            [Θi] Eccentricity, initial inclination imperfections [-].
            Use your own implementation of this value or use the Form5Dot1Imperfections class.
        l_0 : M
            [l0] Effective length of the member, see 5.8.3.2 [m].
        """
        super().__init__()
        self.theta_i = theta_i
        self.l_0 = l_0

    @staticmethod
    def _evaluate(
        theta_i: float,
        l_0: M,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(theta_i=theta_i)
        raise_if_less_or_equal_to_zero(l_0=l_0)
        return theta_i * l_0 / 2



"""Formula 8.10 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_max_curly_brackets, latex_min_curly_brackets
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_negative


class Form8Dot10DesignLapLength(Formula):
    """Class representing formula 8.10 for the calculation of the design lap length :math:`l_{0}` [:math:`mm`]."""

    label = "8.10"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        alpha_1: DIMENSIONLESS,
        alpha_2: DIMENSIONLESS,
        alpha_3: DIMENSIONLESS,
        alpha_5: DIMENSIONLESS,
        alpha_6: DIMENSIONLESS,
        l_b_rqd: MM,
        l_0_min: MM,
    ) -> None:
        """[:math:`l_{0}`] Design lap length [:math:`mm`].

        NEN-EN 1992-1-1+C2:2011 art.8.7.3(1) - Formula (8.10)

        Parameters
        ----------
        alpha_1 : DIMENSIONLESS
            [:math:`α_{1}`] Coefficient for the effect of the form of the bars assuming adequate cover (see figure 8.1) [-].

            :math:`= 1.0` for bars in compression.

            :math:`= 1.0` for straight bars in tension.

            :math:`= 1.0` if :math:`c_{d} <= 3 ⋅ Ø` for bars other than straight in tension (see figure 8.1 (b), (c) and (d)).

            :math:`= 0.7` if :math:`c_{d} > 3 ⋅ Ø` for bars other than straight in tension (see figure 8.1 (b), (c) and (d)).

            Note: see figure 8.3 for values of cd.
        alpha_2 : DIMENSIONLESS
            [:math:`α_{2}`] Coefficient for the effect of minimum concrete cover (see figure 8.3) [-].

            :math:`= 1.0` for bars in compression.

            :math:`= 1 - 0.15 ⋅ (c_{d} - Ø) / Ø <= 1` with a minimum of 0.7 for straight bars in tension.

            :math:`= 1 - 0.15 ⋅ (c_{d} - 3 ⋅ Ø) / Ø <= 1` with a minimum of 0.7 for bars other than straight in tension
            (see figure 8.1 (b), (c) and (d)).

            Note: see figure 8.3 for values of :math:`c_{d}`.
        alpha_3 : DIMENSIONLESS
            [:math:`α_{3}`] Coefficient for the effect of confinement by transverse reinforcement [-].

            :math:`= 1.0` for bars in compression.

            :math:`= 1 - K ⋅ λ <= 1` with a minimum of 0.7 for bars in tension.

            Where: :math:`λ = (ΣA_{st} - ΣA_{st,min}) / A_{s}`.

            Where: :math:`ΣA_{st,min} = A_{s} ⋅ (σ_{sd}/f_{yd})`

            With :math:`A_{s}` = area of one lapped bar [mm²].

            Note: see figure 8.4 for values of :math:`K, A_{s}` and :math:`A_{st}`.
        alpha_5 : DIMENSIONLESS
            [:math:`α_{5}`] Coefficient for the effect of the pressure transverse to the plane of splitting along the design
            anchorage length :math:`l_{bd}` (see 8.6) [-].

            :math:`= 1 - 0.04 ⋅ p <= 1` with a minimum of 0.7 for all types of anchorage in compression.

            Where: p = transverse pressure at ultimate limit state along :math:`l_{bd}` [:math:`MPa`].
        alpha_6 : DIMENSIONLESS
            [:math:`α_{6}`] Coefficient for the effect of reinforcement ratio [-].

            :math:`= (ρ_{1}/25)^{0.5} <= 1.5` with a minimum of 1.0.

            Where: :math:`ρ_{1}` = reinforcement percentage lapped within :math:`0,65 ⋅ l_{0}` from the centre of the lap length
            considered (see figure 8.8) [-].

            Use your own implementation of this formula or use the :class:`SubForm8Dot10Alpha6` class.
        l_b_rqd : MM
            [:math:`l_{b,rqd}`] Basic required anchorage length, for anchoring the force :math:`A_{s} ⋅ σ_{sd}` in a straight bar assuming constant
            bond stress (formula 8.3) [:math:`mm`].

            Use your own implementation for this value or use the :class:`Form8Dot3RequiredAnchorageLength` class.
        l_0_min : MM
            [:math:`l_{0,min}`] Minimum design lap length [:math:`mm`].

            :math:`= max(0.3 ⋅ α_{6} ⋅ l_{b,rqd}, 15 ⋅ Ø, 200)` (formula 8.11).

            Use your own implementation of this formula or use :class:`Form8Dot11MinimumDesignLapLength` class.
        """
        super().__init__()
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.alpha_3 = alpha_3
        self.alpha_5 = alpha_5
        self.alpha_6 = alpha_6
        self.l_b_rqd = l_b_rqd
        self.l_0_min = l_0_min

    @staticmethod
    def _evaluate(
        alpha_1: DIMENSIONLESS,
        alpha_2: DIMENSIONLESS,
        alpha_3: DIMENSIONLESS,
        alpha_5: DIMENSIONLESS,
        alpha_6: DIMENSIONLESS,
        l_b_rqd: MM,
        l_0_min: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_5=alpha_5,
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            l_0_min=l_0_min,
        )
        return max(alpha_1 * alpha_2 * alpha_3 * alpha_5 * alpha_6 * l_b_rqd, l_0_min)

    def latex(self) -> LatexFormula:
        """Returns a LatexFormula representation of the formula."""
        return LatexFormula(
            return_symbol=r"l_{0}",
            result=f"{self:.2f}",
            equation=latex_max_curly_brackets(r"\alpha_1 \cdot \alpha_2 \cdot \alpha_3 \cdot \alpha_5 \cdot \alpha_6 \cdot l_{b,rqd}", r"l_{0,min}"),
            numeric_equation=latex_max_curly_brackets(
                rf"{self.alpha_1:.2f} \cdot {self.alpha_2:.2f} \cdot {self.alpha_3:.2f} \cdot "
                rf"{self.alpha_5:.2f} \cdot {self.alpha_6:.2f} \cdot {self.l_b_rqd:.2f}",
                f"{self.l_0_min:.2f}",
            ),
            comparison_operator_label="=",
        )


class SubForm8Dot10Alpha6(Formula):
    """Class representing the formula for the calculation of the coefficient :math:`α_{6}`."""

    label = "8.8"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, rho_1: DIMENSIONLESS) -> None:
        """[:math:`α_{6}`] Coefficient for the effect of reinforcement ratio [-].

        NEN-EN 1992-1-1+C2:2011 art.8.7.3(1) - Formula (8.8)

        Parameters
        ----------
        rho_1 : DIMENSIONLESS
            [:math:`ρ_{1}`] Reinforcement percentage lapped within :math:`0,65 ⋅ l_{0}` from the centre of the lap length
            considered (see figure 8.8) [-].
        """
        super().__init__()
        self.rho_1 = rho_1

    @staticmethod
    def _evaluate(rho_1: DIMENSIONLESS) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(rho_l=rho_1)
        value_max_1_5 = min((rho_1 / 25) ** 0.5, 1.5)
        return max(value_max_1_5, 1)

    def latex(self) -> LatexFormula:
        """Returns a LatexFormula representation of the formula."""
        argument_1_formula = r"\left(\frac{\rho_1}{25}\right)^{0.5}"
        numerical_argument_1 = rf"\left(\frac{{{self.rho_1:.2f}}}{{25}}\right)^{{0.5}}"
        return LatexFormula(
            return_symbol=r"\alpha_6",
            result=f"{self:.2f}",
            equation=f'{latex_max_curly_brackets(latex_min_curly_brackets(argument_1_formula, "1.5"), "1")}',
            numeric_equation=f'{latex_max_curly_brackets(latex_min_curly_brackets(numerical_argument_1, "1.5"), "1")}',
            comparison_operator_label="=",
        )



"""Latex formula representation."""

from dataclasses import dataclass


@dataclass(frozen=True)
class LatexFormula:
    """Latex formula representation.
    Depending on the context this could include the unit, the formula, the result, etc.

    Attributes
    ----------
    return_symbol: str
        The symbol to return
    result: str
        The result of the formula
    equation: str, default ""
        The formula with symbols
    numeric_equation: str, default ""
        The formula with values (numbers)
    comparison_operator_label: str, default "="
        The label for the comparison operators between the return symbol and the result.
        Could be changed for inequalities.
    """

    return_symbol: str
    result: str
    equation: str = ""
    numeric_equation: str = ""
    comparison_operator_label: str = "="

    @property
    def complete(self) -> str:
        """Complete representation of the formula.

        Returns
        -------
        str
            Return symbol = equation = numeric_equation = result

        """
        all_sub_equations = [
            self.return_symbol,
            self.equation,
            self.numeric_equation,
            self.result,
        ]
        return f" {self.comparison_operator_label} ".join([eq for eq in all_sub_equations if eq != ""])

    @property
    def short(self) -> str:
        """Minimal representation of the formula.

        Returns
        -------
        str
            Return symbol = result

        """
        return f"{self.return_symbol} {self.comparison_operator_label} {self.result}"

    def __str__(self) -> str:
        """String representation of the formula."""
        return self.complete


def latex_fraction(numerator: str | float, denominator: str | float) -> str:
    r"""Return a string which will output: \frac{numerator}{denominator} in latex.

    Examples
    --------
    >>> latex_fraction(1, 2)
    str(\frac{1}{2})

    Parameters
    ----------
    numerator: str | float
        The numerator of the fraction.
    denominator: str | float
        The denominator of the fraction.

    Returns
    -------
    str
        The latex string

    """
    return f"\\frac{{{numerator}}}{{{denominator}}}"


def latex_min_curly_brackets(*args: str | float) -> str:
    r"""Return a string which will output: min{arg_1; arg_2; ...; arg_N} in latex and it will also automatically ensure floats are converted to latex
    text.

    Examples
    --------
    >>> latex_min_curly_brackets(1, 2)
    str(\min \left\{1; 2\right\})

    Parameters
    ----------
    args: str
        The arguments of the min function.

    Returns
    -------
    str
        The latex representation of the min function.
    """
    arguments = [str(arg) for arg in args]
    return f"\\min \\left\\{{{'; '.join(arguments)}\\right\\}}"


def latex_max_curly_brackets(*args: str | float) -> str:
    r"""Return a string which will output: max{arg_1; arg_2; ...; arg_N} in latex and it will also automatically ensure floats are converted to latex
    text.

    Examples
    --------
    >>> latex_max_curly_brackets(1, 2)
    str(\max \left\{1; 2\right\})

    Parameters
    ----------
    args: str
        The arguments of the max function.

    Returns
    -------
    str
        The latex representation of the max function.
    """
    arguments = [str(arg) for arg in args]
    return f"\\max \\left\\{{{'; '.join(arguments)}\\right\\}}"



"""Formula 3.5 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA


class Form3Dot5ApproximationVarianceElasticModulusOverTime(Formula):
    """Class representing formula 3.5 for the approximation of the elastic modulus, Ecm(t) at day t."""

    label = "3.5"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_cm_t: MPA,
        f_cm: MPA,
        e_cm: MPA,
    ) -> None:
        """[Ecm(t)] The approximated elastic modulus at day t [MPa].

        NEN-EN 1992-1-1+C2:2011 art.3.1.3(3) - Formula (3.5)

        Parameters
        ----------
        f_cm_t : MPA
            [fcm(t)] Compressive strength concrete at t days [MPa].
        f_cm : MPA
            [fcm] Average concrete compressive strength on day 28 based on table 3.1 [MPa].
        e_cm : MPA
            [Ecm] Average elastic modulus on day 28 [MPa].

        Returns
        -------
        None
        """
        super().__init__()
        self.f_cm_t = f_cm_t
        self.f_cm = f_cm
        self.e_cm = e_cm

    @staticmethod
    def _evaluate(
        f_cm_t: MPA,
        f_cm: MPA,
        e_cm: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_cm_t < 0:
            raise ValueError(f"Negative f_cm_t: {f_cm_t}. f_cm_t cannot be negative")
        if f_cm < 0:
            raise ValueError(f"Negative f_cm: {f_cm}. f_cm cannot be negative")
        if e_cm < 0:
            raise ValueError(f"Negative e_cm: {e_cm}. e_cm cannot be negative")
        return (f_cm_t / f_cm) ** 0.3 * e_cm



"""Formula 8.11 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_max_curly_brackets
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_negative


class Form8Dot11MinimumDesignLapLength(Formula):
    """Class representing formula 8.11 for the calculation of the minimum design lap length :math:`l_{0,min}` [:math:`mm`]."""

    label = "8.11"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        alpha_6: DIMENSIONLESS,
        l_b_rqd: MM,
        diameter: MM,
    ) -> None:
        """[:math:`l_{0,min}`] Design minimum lap length [:math:`mm`].

        NEN-EN 1992-1-1+C2:2011 art.8.7.3(1) - Formula (8.11)

        Parameters
        ----------
        alpha_6 : DIMENSIONLESS
            [:math:`α_6`] Coefficient for the effect of reinforcement ratio [-].

            :math:`= (ρ_l/25)^{0.5} <= 1.5` with a minimum of 1.0.

            Where: :math:`ρ_l` = reinforcement percentage lapped within :math:`0.65 ⋅ l_0` from the centre of the lap length
            considered (see figure 8.8) [-].

            Use your own implementation for this value or use the :class:`SubForm8Dot10Alpha6` class.
        l_b_rqd: MM
            [:math:`l_{b,rqd}`] Basic required anchorage length, for anchoring the force :math:`A_s ⋅ σ_{sd}` in a straight bar assuming constant
            bond stress (formula 8.3) [:math:`mm`].

            Use your own implementation for this value or use the :class:`Form8Dot3RequiredAnchorageLength` class.
        diameter : MM
            [:math:`Ø`] Diameter of the bar [:math:`mm`].
        """
        super().__init__()
        self.alpha_6 = alpha_6
        self.l_b_rqd = l_b_rqd
        self.diameter = diameter

    @staticmethod
    def _evaluate(
        alpha_6: DIMENSIONLESS,
        l_b_rqd: MM,
        diameter: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            diameter=diameter,
        )
        return max(0.3 * alpha_6 * l_b_rqd, 15 * diameter, 200)

    def latex(self) -> LatexFormula:
        """Returns a representation of the formula in LaTeX format."""
        arg_1_equation = r"0.3 \cdot \alpha_6 \cdot l_{b,rqd}"
        arg_2_equation = r"15 \cdot Ø"
        arg_3_equation = r"200 \ \text{mm}"

        arg_1_numerical_equation = rf"0.3 \cdot {self.alpha_6:.2f} \cdot {self.l_b_rqd:.2f}"
        arg_2_numerical_equation = rf"15 \cdot {self.diameter}"
        arg_3_numerical_equation = r"200"

        return LatexFormula(
            return_symbol=r"l_{0,min}",
            result=f"{self:.2f}",
            equation=f"{latex_max_curly_brackets(arg_1_equation, arg_2_equation, arg_3_equation)}",
            numeric_equation=f"{latex_max_curly_brackets(arg_1_numerical_equation, arg_2_numerical_equation, arg_3_numerical_equation)}",
            comparison_operator_label="=",
        )



"""Formula 9.9 from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailling and specific rules."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import DEG, MM
from blueprints.validations import raise_if_negative


class Form9Dot9MaximumSpacingSeriesOfLinks(Formula):
    """Class representing the formula 9.9 for the calculation of the maximum distance between successive series of links in longitudinal direction
    for slabs.
    """

    label = "9.9"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        d: MM,
        alpha: DEG,
    ) -> None:
        """[smax] Maximum distance between successive series of links in longitudinal direction for slabs [mm].

        NEN-EN 1992-1-1+C2:2011 art.9.3.2(4) - Formula (9.9)

        Parameters
        ----------
        d: MM
            [d] Effective height of the cross-section [mm].
        alpha: DEG
            [α] The angle between the shear reinforcement and the longitudinal axis of the slab (see 9.2.2(1)) [deg].
        """
        super().__init__()
        self.d = d
        self.alpha = alpha

    @staticmethod
    def _evaluate(d: MM, alpha: DEG) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(d=d)

        # Convert the angle from degrees to radians
        alpha_radians = np.deg2rad(alpha)

        # Calculate the cotangent
        cot_alpha = 1 / np.tan(alpha_radians)

        return 0.75 * d * (1 + cot_alpha)



"""Formula 3.1 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA
from blueprints.validations import raise_if_negative


class Form3Dot1EstimationConcreteCompressiveStrength(Formula):
    """Class representing formula 3.1 for the estimation of the concrete compressive strength, fcm(t),  after t days
    with an average temperature of 20 degrees Celsius.
    """

    label = "3.1"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        beta_cc_t: float,
        f_cm: MPA,
    ) -> None:
        """[fcm(t)] The estimated concrete compressive strength [MPa].

        NEN-EN 1992-1-1+C2:2011 art.3.1.2(6) - Formula (3.1)

        Parameters
        ----------
        beta_cc_t : float
            [βcc(t)] Coefficient dependent of the age of concrete [-].
        f_cm : MPA
            [fcm] Average concrete compressive strength on day 28 based on table 3.1 [MPa].

        Returns
        -------
        None
        """
        super().__init__()
        self.beta_cc_t = beta_cc_t
        self.f_cm = f_cm

    @staticmethod
    def _evaluate(
        beta_cc_t: float,
        f_cm: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(beta_cc_t=beta_cc_t, f_cm=f_cm)
        return beta_cc_t * f_cm



"""Formula 9.10 from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailling and specific rules."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative


class Form9Dot10MaximumSpacingBentUpBars(Formula):
    """Class representing the formula 9.10 for the calculation of the maximum longitudinal spacing of bent up bars for slabs."""

    label = "9.10"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, d: MM) -> None:
        """[smax] Maximum longitudinal spacing of bent up bars for slabs [mm].

        NEN-EN 1992-1-1+C2:2011 art.9.3.2(4) - Formula (9.10)

        Parameters
        ----------
        d: MM
            [d] Effective height of the cross-section [mm].
        """
        super().__init__()
        self.d = d

    @staticmethod
    def _evaluate(d: MM) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(d=d)
        return d



"""Formula 7.3 from NEN-EN 1992-1-1+C2:2011: Chapter 7 - Serviceability limit state (SLS)."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import KN, MM2, MPA
from blueprints.unit_conversion import KN_TO_N


class Form7Dot3CoefficientKc(Formula):
    """Class representing the formula 7.3 for the coefficient kc for flanges of tubular cross-sections and T-sections."""

    label = "7.3"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_cr: KN,
        a_ct: MM2,
        f_ct_eff: MPA,
    ) -> None:
        """[kc] Calculates kc for flanges of tubular cross-sections and T-sections [-].

        NEN-EN 1992-1-1:2011 art.7.3.2(2) - Formula (7.3)

        Parameters
        ----------
        f_cr : KN
            [Fcr] Absolute value of the tensile force within the flange immediately before cracking due to the cracking moment calculated with
            fct,eff [kN].
        a_ct : MM2
            [Act] Area of the concrete within the tension zone. The tension zone is that part of the cross-section that, according to the calculation,
            is under tension just before the first crack occurs [mm²].
        f_ct_eff : MPA
            [fc,eff] Average value of the tensile strength of the concrete at the time when the first cracks can be expected [MPa].
        """
        super().__init__()
        self.f_cr = f_cr
        self.a_ct = a_ct
        self.f_ct_eff = f_ct_eff

    @staticmethod
    def _evaluate(
        f_cr: KN,
        a_ct: MM2,
        f_ct_eff: MPA,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if a_ct <= 0:
            raise ValueError("The value of a_ct must be greater than zero.")
        if f_ct_eff <= 0:
            raise ValueError("The value of f_ct_eff must be greater than zero.")
        return max(0.9 * (abs(f_cr) * KN_TO_N / (a_ct * f_ct_eff)), 0.5)



"""Formula 3.7 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula


class Form3Dot7NonLinearCreepCoefficient(Formula):
    """Class representing formula 3.7 for the calculation of the non-linear creep coefficient."""

    label = "3.7"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        phi_inf_t0: float,
        k_sigma: float,
    ) -> None:
        """[φnl(∞,t0)] The non-linear creep coefficient [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(4) - Formula (3.7)

        Parameters
        ----------
        phi_inf_t0 : float
            [φ(∞, t0)] Creep coefficient if high accuracy is not required use figure 3.1 and/or use appendix B [-].
        k_sigma : float
            [kσ] Stress-strength ratio (σc / fck(t0)) [-].

        Returns
        -------
        None
        """
        super().__init__()
        self.phi_inf_t0 = phi_inf_t0
        self.k_sigma = k_sigma

    @staticmethod
    def _evaluate(
        phi_inf_t0: float,
        k_sigma: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if phi_inf_t0 < 0:
            raise ValueError(f"Negative phi_inf_t0: {phi_inf_t0}. phi_inf_t0 cannot be negative")
        if k_sigma < 0:
            raise ValueError(f"Negative k_sigma: {k_sigma}. k_sigma cannot be negative")
        return phi_inf_t0 * np.exp(1.5 * (k_sigma - 0.45))



"""Formula 6.1 from NEN-EN 1992-1-1+C2:2011: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import KN


class Form6Dot1DesignShearStrength(Formula):
    """Class representing formula 6.1 for the design shear strength, VRd."""

    label = "6.1"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        v_rd_s: KN,
        v_ccd: KN,
        v_td: KN,
    ) -> None:
        """[VRd] Design shear resistance of an element with shear reinforcement.

        NEN-EN 1992-1-1+C2:2011 art.6.2.1(2) - Formula (6.1)

        Parameters
        ----------
        v_rd_s : KN
            [VRd,s] Design shear resistance of an element with shear reinforcement [kN].
        v_ccd : KN
            [Vccd] Design value of the shear force component in the compression area in case of a change in height [kN].
        v_td : KN
            [Vtd] Design value of the shear force component of the tensile force in the reinforcement in case of a change in height [kN].
        """
        super().__init__()
        self.v_rd_s = v_rd_s
        self.v_ccd = v_ccd
        self.v_td = v_td

    @staticmethod
    def _evaluate(
        v_rd_s: KN,
        v_ccd: KN,
        v_td: KN,
    ) -> KN:
        """Evaluates the formula, for more information see the __init__ method."""
        if v_rd_s < 0:
            raise ValueError(f"Negative v_rd_s: {v_rd_s}. v_rd_s cannot be negative")
        if v_ccd < 0:
            raise ValueError(f"Negative v_ccd: {v_ccd}. v_ccd cannot be negative")
        if v_td < 0:
            raise ValueError(f"Negative v_td: {v_td}. v_td cannot be negative")
        return v_rd_s + v_ccd + v_td



"""Formula 9.14 from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailing and specific rules."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import KN, MM
from blueprints.validations import raise_if_negative


class Form9Dot14SplittingForceColumnOnRock(Formula):
    """Class representing the formula 9.14 for the calculation of the splitting force on a column footing on rock."""

    label = "9.14"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        c: MM,
        h: MM,
        n_ed: KN,
    ) -> None:
        """[Fs] Splitting force on a column footing on rock [kN].

        NEN-EN 1992-1-1+C2:2011 art.9.8.4(2) - Formula (9.14)

        Parameters
        ----------
        c: MM
            [c] Width over which NEd is applied [mm].
        h: MM
            [h] Lesser of b and H from figure 9.14 [mm].
        n_ed: KN
            [NEd] Design value of the applied axial force [kN].
        """
        super().__init__()
        self.c = c
        self.h = h
        self.n_ed = n_ed

    @staticmethod
    def _evaluate(
        c: MM,
        h: MM,
        n_ed: KN,
    ) -> KN:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(c=c, h=h, n_ed=n_ed)
        return 0.25 * (1 - (c / h)) * n_ed



"""Formula 3.26 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA


class Form3Dot26IncreasedStrainAtMaxStrength(Formula):
    """Class representing formula 3.26 for the calculation of the increased strain at the maximum strength due to enclosed concrete."""

    label = "3.26"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_ck: MPA,
        f_ck_c: MPA,
        epsilon_c2: float,
    ) -> None:
        """[εc2,c] Increased strain at the maximum strength due to enclosed concrete. [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.9(2) - Formula (3.26)

        Parameters
        ----------
        f_ck : MPA
            [fck] Characteristic compressive strength [MPa]
        f_ck_c : MPA
            [fck,c] Increased characteristic compressive strength due to enclosed concrete [MPa].
            See classes Form3Dot24IncreasedCharacteristicCompressiveStrength and/or Form3Dot25IncreasedCharacteristicCompressiveStrength
        epsilon_c2 : float
            [εc2] Strain at maximum strength [-]

        Returns
        -------
        None
        """
        super().__init__()
        self.f_ck = f_ck
        self.f_ck_c = f_ck_c
        self.epsilon_c2 = epsilon_c2

    @staticmethod
    def _evaluate(
        f_ck: MPA,
        f_ck_c: MPA,
        epsilon_c2: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_ck < 0:
            raise ValueError(f"Invalid f_ck: {f_ck}. f_ck cannot be negative")
        if f_ck_c < 0:
            raise ValueError(f"Invalid f_ck_c: {f_ck_c}. f_ck_c cannot be negative")
        return epsilon_c2 * (f_ck_c / f_ck) ** 2



"""Formula 3.30 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import HOURS, PERCENTAGE


class Form3Dot30RatioLossOfPreStressClass3(Formula):
    """Class representing formula 3.30 for the calculation of the ratio between loss of pre-stress and initial pre-stress of class 3."""

    label = "3.30"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        rho_1000: PERCENTAGE,
        mu: float,
        t: HOURS,
    ) -> None:
        """[Δσpr / σpi] Ratio between loss of pre-stress and initial pre-stress for class 3. [-].

        NEN-EN 1992-1-1+C2:2011 art.3.3.2(7) - Formula (3.30)

        Parameters
        ----------
        rho_1000 : PERCENTAGE
            [ρ1000] Value of relaxation loss at 1000h after prestressing at an average temperature of 20 degrees Celsius [%]
        mu : float
            [μ] Ratio between initial pre-stress and characteristic tensile strength [-]
            = σpi / fpk
            Use your own implementation of this formula or use sub_formula_3_28_39_30 class SubForm3Dot282930Mu.
        t : HOURS
            [t] Time after prestressing [hours]

        Returns
        -------
        None
        """
        super().__init__()
        self.rho_1000 = rho_1000
        self.mu = mu
        self.t = t

    @staticmethod
    def _evaluate(
        rho_1000: PERCENTAGE,
        mu: float,
        t: HOURS,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if rho_1000 < 0:
            raise ValueError(f"Invalid rho_1000: {rho_1000}. rho_1000 cannot be negative")
        if t < 0:
            raise ValueError(f"Invalid t: {t}. t cannot be negative")
        return 1.98 * rho_1000 * np.exp(8 * mu) * (t / 1000) ** (0.75 * (1 - mu)) * 10**-5



"""Formula 9.16 from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailling and specific rules."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import KN, KN_M, M
from blueprints.validations import raise_if_negative


class Form9Dot16MinimumForceOnInternalBeamLine(Formula):
    """Class representing the formula 9.16 for calculating the minimum force on an internal beam line for floors without screeds."""

    label = "9.16"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        q_3: KN_M,
        l_1: M,
        l_2: M,
        q_4: KN,
    ) -> None:
        """[Ftie] Minimum force on an internal beam line [kN].

        NEN-EN 1992-1-1+C2:2011 art.9.10.2.3(4) - Formula (9.16)

        Parameters
        ----------
        q_3: KN_M
            [q3] May be found in national annex, recommended value is 20 [kN/m].
        l_1: M
            [l1] span length of floor slabs on either side of the beam, see figure 9.15 [m].
        l_2: M
            [l2] span length of floor slabs on either side of the beam, see figure 9.15 [m].
        q_4: KN
            [Q4] May be found in national annex, recommended value is 70 [kN].
        """
        super().__init__()
        self.q_3 = q_3
        self.l_1 = l_1
        self.l_2 = l_2
        self.q_4 = q_4

    @staticmethod
    def _evaluate(
        q_3: KN_M,
        l_1: M,
        l_2: M,
        q_4: KN,
    ) -> KN:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(q_3=q_3, l_1=l_1, l_2=l_2, q_4=q_4)
        return max(q_3 * (l_1 + l_2) / 2, q_4)



"""
Pytest tests for validation functions in Blueprints.

It includes tests for:
- raise_if_less_or_equal_to_zero: Ensuring it raises an exception for non-positive values.
- raise_if_negative: Ensuring it raises an exception for negative values.
"""

import pytest

from blueprints.validations import (
    GreaterThan90Error,
    LessOrEqualToZeroError,
    NegativeValueError,
    raise_if_greater_than_90,
    raise_if_less_or_equal_to_zero,
    raise_if_negative,
)


def test_raise_if_less_or_equal_to_zero_with_positive_values() -> None:
    """Test that no exception is raised for positive values."""
    raise_if_less_or_equal_to_zero(a=1, b=2, c=3)


def test_raise_if_less_or_equal_to_zero_with_zero() -> None:
    """Test that NonPositiveValueError is raised for zero values."""
    with pytest.raises(LessOrEqualToZeroError):
        raise_if_less_or_equal_to_zero(a=0)


def test_raise_if_less_or_equal_to_zero_with_negative() -> None:
    """Test that NonPositiveValueError is raised for negative values."""
    with pytest.raises(LessOrEqualToZeroError):
        raise_if_less_or_equal_to_zero(a=-1)


def test_raise_if_negative_with_positive_values() -> None:
    """Test that no exception is raised for positive values."""
    raise_if_negative(a=1, b=2, c=3)


def test_raise_if_negative_with_zero() -> None:
    """Test that no exception is raised for zero values."""
    raise_if_negative(a=0)


def test_raise_if_negative_with_negative() -> None:
    """Test that NegativeValueError is raised for negative values."""
    with pytest.raises(NegativeValueError):
        raise_if_negative(a=-1)


def test_raise_if_greater_90_with_less_than_90() -> None:
    """Test that no exception is raised for values lesser than 90."""
    raise_if_greater_than_90(a=-10, b=0, c=85)


def test_raise_if_greater_90_with_90() -> None:
    """Test that no exception is raised for 90."""
    raise_if_greater_than_90(a=90)


def test_raise_if_greater_90_with_greater_than_90() -> None:
    """Test that GreaterThan90Error is raised for values greater than 90."""
    with pytest.raises(GreaterThan90Error):
        raise_if_greater_than_90(a=95)



"""Formula 3.6 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA


class Form3Dot6CreepDeformationOfConcrete(Formula):
    """Class representing formula 3.6 for the calculation of creep deformation of concrete."""

    label = "3.6"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        phi_inf_t0: float,
        sigma_c: MPA,
        e_c: MPA,
    ) -> None:
        """εcc(∞,t0) Creep deformation of concrete at the time t = ∞ for a constant concrete compressive stress σc applied at time t0 [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(3) - Formula (3.6)

        Parameters
        ----------
        phi_inf_t0 : float
            [φ(∞, t0)] Creep coefficient if high accuracy is not required use figure 3.1 else use appendix B [-].
        sigma_c : MPA
            [σc] Concrete compressive stress [MPa].
        e_c : MPA
            [Ec] tangent modulus = 1.05 * Ecm. According to art.3.1.4(2) [MPa].

        Returns
        -------
        None
        """
        super().__init__()
        self.phi_inf_t0 = phi_inf_t0
        self.sigma_c = sigma_c
        self.e_c = e_c

    @staticmethod
    def _evaluate(
        phi_inf_t0: float,
        sigma_c: MPA,
        e_c: MPA,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if phi_inf_t0 < 0:
            raise ValueError(f"Negative phi_inf_t0: {phi_inf_t0}. phi_inf_t0 cannot be negative")
        if sigma_c < 0:
            raise ValueError(f"Negative sigma_c: {sigma_c}. sigma_c cannot be negative")
        if e_c < 0:
            raise ValueError(f"Negative e_c: {e_c}. e_c cannot be negative")
        return phi_inf_t0 * sigma_c / e_c



"""Formula 4.1 from NEN-EN 1992-1-1+C2:2011: Chapter 4 - Durability and cover to reinforcement."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative


class Form4Dot1NominalConcreteCover(Formula):
    """Class representing the formula 4.1 for the calculation of the nominal concrete cover :math:`c_{nom}` [:math:`mm`]."""

    label = "4.1"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        c_min: MM,
        delta_c_dev: MM,
    ) -> None:
        """[:math:`c_{nom}`] Calculates the nominal concrete cover [:math:`mm`].

        NEN-EN 1992-1-1+C2:2011 art.4.4.1.1 (2) - Formula (4.1)

        Parameters
        ----------
        c_min: MM
            [:math:`c_{min}`] Minimum concrete cover based on art. 4.4.1.2 [:math:`mm`].
        delta_c_dev: MM
            [:math:`Δc_{dev}`] Construction tolerance based on art. 4.4.1.3 [:math:`mm`].
        """
        super().__init__()
        self.c_min = c_min
        self.delta_c_dev = delta_c_dev

    @staticmethod
    def _evaluate(
        c_min: MM,
        delta_c_dev: MM,
    ) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(c_min=c_min, delta_c_dev=delta_c_dev)
        return c_min + delta_c_dev

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 4.1."""
        return LatexFormula(
            return_symbol=r"c_{nom}",
            result=str(self),
            equation=r"c_{min}+\Delta c_{dev}",
            numeric_equation=f"{self.c_min}+{self.delta_c_dev}",
            comparison_operator_label="=",
        )



"""Formula 4.2 from NEN-EN 1992-1-1+C2:2011: Chapter 4 - Durability and cover to reinforcement."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_max_curly_brackets
from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative


class Form4Dot2MinimumConcreteCover(Formula):
    """Class representing the formula 4.2 for the calculation of the minimum concrete cover :math:`c_{min}` [:math:`mm`]."""

    label = "4.2"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        c_min_b: MM,
        c_min_dur: MM,
        delta_c_dur_gamma: MM,
        delta_c_dur_st: MM,
        delta_c_dur_add: MM,
    ) -> None:
        """[:math:`c_{min}`] Calculates the minimum concrete cover [:math:`mm`].

        A minimum concrete cover of 10 mm is required, even if the calculated value is lower.

        NEN-EN 1992-1-1+C2:2011 art.4.4.1.2 (2) - formula (4.2)

        Parameters
        ----------
        c_min_b: MM
            [:math:`c_{min,b}`] The minimum concrete cover based on the adhesion requirements based on art. 4.4.1.2 (3) [:math:`mm`].
        c_min_dur: MM
            [:math:`c_{min,dur}`] The minimum concrete cover based on environmental conditions based on art. 4.4.1.2 (5) [:math:`mm`].
        delta_c_dur_gamma: MM
            [:math:`Δc_{dur,γ}`] An additional safety requirement based on art. 4.4.1.2 (6) [:math:`mm`].
        delta_c_dur_st: MM
            [:math:`Δc_{dur,st}`] A reduction of minimum concrete cover when using stainless steel based on art. 4.4.1.2 (7) [:math:`mm`].
        delta_c_dur_add: MM
            [:math:`Δc_{dur,add}`] A reduction of minimum concrete cover when using additional protection based on art. 4.4.1.2 (8) [:math:`mm`].
        """
        super().__init__()
        self.c_min_b = c_min_b
        self.c_min_dur = c_min_dur
        self.delta_c_dur_gamma = delta_c_dur_gamma
        self.delta_c_dur_st = delta_c_dur_st
        self.delta_c_dur_add = delta_c_dur_add

    @staticmethod
    def _evaluate(
        c_min_b: MM,
        c_min_dur: MM,
        delta_c_dur_gamma: MM,
        delta_c_dur_st: MM,
        delta_c_dur_add: MM,
    ) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(
            c_min_b=c_min_b,
            c_min_dur=c_min_dur,
            delta_c_dur_gamma=delta_c_dur_gamma,
            delta_c_dur_st=delta_c_dur_st,
            delta_c_dur_add=delta_c_dur_add,
        )
        minimum_cover = 10  # mm
        return max(c_min_b, c_min_dur + delta_c_dur_gamma - delta_c_dur_st - delta_c_dur_add, minimum_cover)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 4.2."""
        return LatexFormula(
            return_symbol=r"c_{min}",
            result=str(self),
            equation=latex_max_curly_brackets(
                r"c_{min,b}",
                r"c_{min,dur}+\Delta c_{dur,\gamma}-\Delta c_{dur,st}-\Delta c_{dur,add}",
                r"10 \ \text{mm}",
            ),
            numeric_equation=latex_max_curly_brackets(
                self.c_min_b,
                f"{self.c_min_dur}+{self.delta_c_dur_gamma}-{self.delta_c_dur_st}-{self.delta_c_dur_add}",
                10,
            ),
            comparison_operator_label="=",
        )



"""Formula 8.9 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import KN, MM, MM2, MPA
from blueprints.unit_conversion import N_TO_KN
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form8Dot9AnchorageCapacityWeldedTransverseBarSmallDiameter(Formula):
    """Class representing the formula 8.9 for the calculation of the anchorage capacity of a welded cross bar for nominal bar diameters smaller
    than 12 mm.
    """

    label = "8.9"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_wd: KN,
        diameter_t: MM,
        diameter_l: MM,
        a_s: MM2,
        f_cd: MPA,
    ) -> None:
        """[:math:`F_{btd}`] Anchorage capacity of a welded cross bar for nominal bar diameters smaller than 12 mm [:math:`kN`].

        NEN-EN 1992-1-1+C2:2011 art.8.6(5) - formula 8.9

        Parameters
        ----------
        f_wd : KN
            [:math:`F_{wd}`] Design shear strength of weld (specified as a factor times :math:`A_{s} ⋅ f_{yd}`; say :math:`0.5 ⋅ A_{s} ⋅ f_{yd}`
            where :math:`A_{s}` is the cross-section of the anchored bar and :math:`f_{yd}` is its design yield strength)  [:math:`kN`].
        diameter_t : MM
            [:math:`ø_{t}`] Diameter of the transverse bar [:math:`mm`].

            Note: :math:`ø_{t} =< 12 mm`.
        diameter_l : MM
            [:math:`ø_{l}`] Diameter of the bar to be anchored [:math:`mm`].

            Note: :math:`ø_{l} =< 12 mm`.
        a_s : MM2
            [:math:`A_{s}`] Cross-section of the anchored bar [:math:`mm^{2}`].
        f_cd : MPA
            [:math:`f_{cd}`] Design compressive strength of concrete [:math:`MPa`].
        """
        super().__init__()
        self.f_wd = f_wd
        self.diameter_t = diameter_t
        self.diameter_l = diameter_l
        self.a_s = a_s
        self.f_cd = f_cd

    @staticmethod
    def _evaluate(
        f_wd: KN,
        diameter_t: MM,
        diameter_l: MM,
        a_s: MM2,
        f_cd: MPA,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            f_wd=f_wd,
            diameter_t=diameter_t,
            a_s=a_s,
            f_cd=f_cd,
        )
        raise_if_less_or_equal_to_zero(diameter_l=diameter_l)
        return min(f_wd, N_TO_KN * 16 * a_s * f_cd * (diameter_t / diameter_l))



"""Formula 3.14 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA


class Form3Dot14StressStrainForShortTermLoading(Formula):
    """Class representing formula 3.14, which calculates the compressive stress-strength ratio."""

    source_document = NEN_EN_1992_1_1_C2_2011
    label = "3.14"

    def __init__(
        self,
        k: float,
        eta: float,
    ) -> None:
        """[σc / fcm] Compressive stress-strength ratio [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.5(1) - Formula (3.14)

        Parameters
        ----------
        k : float
            [k] [-].
            = 1.05 * Ecm * |εc1| / fcm
            Use your own implementation of this formula or use the SubForm3Dot14K class.
        eta : float
            [η] Strain - peak-strain ratio [-].
            = εc / εc1
            Use your own implementation of this formula or use the SubForm3Dot14Eta class.

        Returns
        -------
        None
        """
        super().__init__()
        self.k = k
        self.eta = eta

    @staticmethod
    def _evaluate(
        k: float,
        eta: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if k < 0:
            raise ValueError(f"Invalid k: {k}. k cannot be negative")
        if eta < 0:
            raise ValueError(f"Invalid eta: {eta}. eta cannot be negative")
        return (k * eta - eta**2) / (1 + (k - 2) * eta)


class SubForm3Dot14Eta(Formula):
    """Class representing sub-formula 1 for formula 3.14, which calculates eta."""

    source_document = NEN_EN_1992_1_1_C2_2011
    label = "3.14"

    def __init__(
        self,
        epsilon_c: float,
        epsilon_c1: float,
    ) -> None:
        """[η] Strain - peak-strain ratio [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.5(1) - η

        Parameters
        ----------
        epsilon_c : float
            [εc] Strain concrete [-].
        epsilon_c1 : float
            [εc1] Strain concrete at peak-stress following table 3.1 [-].
        """
        super().__init__()
        self.epsilon_c = epsilon_c
        self.epsilon_c1 = epsilon_c1

    @staticmethod
    def _evaluate(
        epsilon_c: float,
        epsilon_c1: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        return epsilon_c / epsilon_c1


class SubForm3Dot14K(Formula):
    """Class representing sub-formula 2 for formula 3.14, which calculates k."""

    source_document = NEN_EN_1992_1_1_C2_2011
    label = "3.14"

    def __init__(self, e_cm: MPA, epsilon_c1: float, f_cm: MPA) -> None:
        """[k] [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.5(1) - k

        Parameters
        ----------
        e_cm : MPA
            [Ecm] Elastic modulus concrete [MPa].
        epsilon_c1 : float
            [εc1] Strain concrete at peak-stress following table 3.1 [-].
        f_cm : MPA
            [fcm] Compressive strength concrete [MPa].
        """
        super().__init__()
        self.e_cm = e_cm
        self.epsilon_c1 = epsilon_c1
        self.f_cm = f_cm

    @staticmethod
    def _evaluate(
        e_cm: MPA,
        epsilon_c1: float,
        f_cm: MPA,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if e_cm < 0:
            raise ValueError(f"Invalid e_cm: {e_cm}. e_cm cannot be negative")
        if f_cm <= 0:
            raise ValueError(f"Invalid f_cm: {f_cm}. f_cm cannot be negative or zero")
        return 1.05 * e_cm * abs(epsilon_c1) / f_cm



"""Formula 8.8N from NEN-EN 1992-1-1+C2:2011: Chapter 8 - Detailing of reinforcement and prestressing tendons."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import DIMENSIONLESS, KN, MM, MPA
from blueprints.unit_conversion import N_TO_KN
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form8Dot8nAnchorageCapacityWeldedTransverseBar(Formula):
    """Class representing the formula 8.8N for the calculation of the anchorage capacity of welded transverse bar, welded on the inside of the main
    bar.
    """

    label = "8.8N"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        l_td: MM,
        diameter_t: MM,
        sigma_td: MPA,
        f_wd: KN,
    ) -> None:
        """[:math:`F_{btd}`] Anchorage capacity of welded transverse bar, welded on the inside of the main bar [:math:`kN`].

            Note: Value may be found in National Annex.

        NEN-EN 1992-1-1+C2:2011 art.8.6(2) - formula (8.8N)

        Parameters
        ----------
        l_td: MM
            [:math:`l_{td}`] Design length of transverse bar [:math:`mm`].

            :math:`= 1.16 ⋅ ø_{t} ⋅ (f_{yd}/σ_{td})^{0.5} ≤ l_{t}`

            Use your own implementation of this formula or use the SubForm8Dot8nDesignLengthOfTransverseBar class.
        diameter_t: MM
            [:math:`ø_{t}`] Diameter of transverse bar [:math:`mm`].
        sigma_td: MPA
            [:math:`σ_{td}`] Concrete stress [:math:`MPa`].

            :math:`=(f_{ctd}+σ_{cm})/y ≤ 3⋅f_{cd}`

            Use your own implementation of this formula or use the SubForm8Dot8nConcreteStress class.
        f_wd: KN
            [:math:`F_{wd}`] Design shear strength of weld (specified as a factor times :math:`A_{s}⋅f_{yd}`; say :math:`0.5⋅A_{s}⋅f_{yd}` where
            :math:`A_{s}` is the cross-section of the anchored bar and fyd is its design yield strength)  [kN].
        """
        super().__init__()
        self.l_td = l_td
        self.diameter_t = diameter_t
        self.sigma_td = sigma_td
        self.f_wd = f_wd

    @staticmethod
    def _evaluate(
        l_td: MM,
        diameter_t: MM,
        sigma_td: MPA,
        f_wd: KN,
    ) -> KN:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(
            l_td=l_td,
            diameter_t=diameter_t,
            sigma_td=sigma_td,
            f_wd=f_wd,
        )
        return min(l_td * diameter_t * sigma_td * N_TO_KN, f_wd)


class SubForm8Dot8nDesignLengthOfTransverseBar(Formula):
    """Class representing sub-formula for formula 8.8N, which calculates the design length of the transverse bar."""

    label = "8.8N"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        diameter_t: MM,
        f_yd: MPA,
        sigma_td: MPA,
        l_t: MM,
    ) -> None:
        """[:math:`l_{td}`] Design length of transverse bar [:math:`mm`].

        NEN-EN 1992-1-1+C2:2011 art.8.6(2) - :math:`l_{td}`

        Parameters
        ----------
        diameter_t: MM
            [:math:`ø_{t}`] Diameter of transverse bar [:math:`mm`].
        f_yd: MPA
            [:math:`f_{yd}`] Design yield strength of bar [:math:`MPa`].
        sigma_td: MPA
            [:math:`σ_{td}`] Concrete stress [:math:`MPa`].

            :math:`=(f_{ctd}+σ_{cm})/y ≤ 3⋅f_{cd}`

            Use your own implementation of this formula or use the SubForm8Dot8nConcreteStress class.
        l_t: MM
            [:math:`l_{t}`] Length of transverse bar, but not more than the spacing of bars to be anchored [:math:`mm`].
        """
        super().__init__()
        self.diameter_t = diameter_t
        self.f_yd = f_yd
        self.sigma_td = sigma_td
        self.l_t = l_t

    @staticmethod
    def _evaluate(
        diameter_t: MM,
        f_yd: MPA,
        sigma_td: MPA,
        l_t: MM,
    ) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(
            diameter_t=diameter_t,
            f_yd=f_yd,
            l_t=l_t,
        )
        raise_if_less_or_equal_to_zero(sigma_td=sigma_td)
        return min(1.16 * diameter_t * (f_yd / sigma_td) ** 0.5, l_t)


class SubForm8Dot8nConcreteStress(Formula):
    """Class representing sub-formula for formula 8.8N, which calculates the concrete stress."""

    label = "8.8N"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_ctd: MPA,
        sigma_cm: MPA,
        y_function: DIMENSIONLESS,
        f_cd: MPA,
    ) -> None:
        """[:math:`σ_{td}`] Concrete stress [:math:`MPa`].

        NEN-EN 1992-1-1+C2:2011 art.8.6(2) - σtd

        Parameters
        ----------
        f_ctd: MPA
            [:math:`f_{ctd}`] Design tensile strength of concrete [:math:`MPa`].
        sigma_cm: MPA
            [:math:`σ_{cm}`] Compression in the concrete perpendicular to both bars (mean value) [:math:`MPa`].
        y_function: DIMENSIONLESS
            [y] A function [-]

            :math:`= 0.015 + 0.14 ⋅ exp(-0.18⋅x)`

            Use your own implementation of this formula or use the SubForm8Dot8nFunctionY class.
        f_cd: MPA
            [:math:`f_{cd}`] Design value compressive strength of concrete [:math:`MPa`].
        """
        super().__init__()
        self.f_ctd = f_ctd
        self.sigma_cm = sigma_cm
        self.y_function = y_function
        self.f_cd = f_cd

    @staticmethod
    def _evaluate(
        f_ctd: MPA,
        sigma_cm: MPA,
        y_function: DIMENSIONLESS,
        f_cd: MPA,
    ) -> MPA:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(
            f_ctd=f_ctd,
            sigma_cm=sigma_cm,
            f_cd=f_cd,
        )
        raise_if_less_or_equal_to_zero(y_function=y_function)
        return min((f_ctd + sigma_cm) / y_function, 3 * f_cd)


class SubForm8Dot8nFunctionY(Formula):
    """Class representing sub-formula for formula 8.8N, which calculates the function y."""

    label = "8.8N"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        x_function: DIMENSIONLESS,
    ) -> None:
        """[y] A function [-].

        NEN-EN 1992-1-1+C2:2011 art.8.6(2) - y

        Parameters
        ----------
        x_function: DIMENSIONLESS
            [x] A function accounting for the geometry [-]

            :math:`= 2⋅(c/ø_{t}) + 1`

            Use your own implementation of this formula or use the SubForm8Dot8nFunctionX class.
        """
        super().__init__()
        self.x_function = x_function

    @staticmethod
    def _evaluate(
        x_function: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(x_function=x_function)
        return 0.015 + 0.14 * np.exp(-0.18 * x_function)


class SubForm8Dot8nFunctionX(Formula):
    """Class representing sub-formula for formula 8.8N, which calculates the function x."""

    label = "8.8N"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        cover: MM,
        diameter_t: MM,
    ) -> None:
        """[x] A function accounting for the geometry [-].

        NEN-EN 1992-1-1+C2:2011 art.8.6(2) - x

        Parameters
        ----------
        cover: MM
            [c] Concrete cover perpendicular to both bars [:math:`mm`].
        diameter_t: MM
            [:math:`ø_{t}`] Diameter of transverse bar [:math:`mm`].
        """
        super().__init__()
        self.cover = cover
        self.diameter_t = diameter_t

    @staticmethod
    def _evaluate(
        cover: MM,
        diameter_t: MM,
    ) -> DIMENSIONLESS:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(cover=cover)
        raise_if_less_or_equal_to_zero(diameter_t=diameter_t)
        return 2 * (cover / diameter_t) + 1



"""Formula 3.24 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA


class Form3Dot24IncreasedCharacteristicCompressiveStrength(Formula):
    """Class representing formula 3.24 for the calculation of the increased characteristic compressive strength due to enclosed concrete."""

    label = "3.24"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_ck: MPA,
        sigma_2: MPA,
    ) -> None:
        """[fck,c] Increased characteristic compressive strength due to enclosed concrete [MPa].

        NEN-EN 1992-1-1+C2:2011 art.3.1.9(2) - Formula (3.24)

        Parameters
        ----------
        f_ck : MPA
            [fck] Characteristic compressive strength [MPa]
        sigma_2 : MPA
            [σ2] Effective compressive stress in transverse direction [MPa]
            σ2 <= 0.05 * fck

        Returns
        -------
        None
        """
        super().__init__()
        self.f_ck = f_ck
        self.sigma_2 = sigma_2

    @staticmethod
    def _evaluate(
        f_ck: MPA,
        sigma_2: MPA,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_ck < 0:
            raise ValueError(f"Invalid f_ck: {f_ck}. f_ck cannot be negative")
        if sigma_2 > 0.05 * f_ck:
            raise ValueError(f"Invalid sigma_2: {sigma_2}. Try using Form3Dot25IncreasedCharacteristicCompressiveStrength class.")
        return f_ck * (1.000 + 5.0 * sigma_2 / f_ck)



"""Formula 2.1a from NEN 9997-1+C2:2017: Chapter 2: Basis of geotechnical design."""

from blueprints.codes.eurocode.nen_9997_1_c2_2017 import NEN_9997_1_C2_2017
from blueprints.codes.formula import Formula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_negative


class Form2Dot1aDesignValueLoad(Formula):
    """Class representing formula 2.1a for the calculation of the design value :math:`F_{d}` of actions."""

    label = "2.1a"
    source_document = NEN_9997_1_C2_2017

    def __init__(self, gamma_f: DIMENSIONLESS, f_rep: float) -> None:
        """[:math:`F_{d}`] Design value of actions.

        NEN 9997-1+C2:2017 art.2.4.6.1(2) - (Formula 2.1a)

        Parameters
        ----------
        gamma_f : DIMENSIONLESS
            [:math:`γ_{F}`] partial factor for actions for persistent and transient situations defined in annex A [-].
        f_rep : float
            [:math:`F_{rep}`] Representative value of actions.

            Use your own implementation for this value or use :class:`Form2Dot1bRepresentativeValue`.
        """
        super().__init__()
        self.gamma_f = gamma_f
        self.f_rep = f_rep

    @staticmethod
    def _evaluate(
        gamma_f: DIMENSIONLESS,
        f_rep: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(gamma_f=gamma_f)
        return gamma_f * f_rep



"""Formula 9.3 from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailing of members and particular rules."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import KN, MM
from blueprints.validations import raise_if_negative


class Form9Dot3ShiftInMomentDiagram(Formula):
    """Class representing the formula 9.3 for the calculation of anchorage length of bottom reinforcement at an end support using the shift rule."""

    label = "9.3"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        v_ed: KN,
        a_l: MM,
        z: MM,
        n_ed: KN,
    ) -> None:
        """[FEd] Force to be anchored according to the shift rule [kN].

        NEN-EN 1992-1-1+C2:2011 art.9.2.1.4(2) - Formula (9.3)

        Parameters
        ----------
        v_ed: KN
            [VEd] Design value shear force [kN].
        a_l: MM
            [al] Shift in the moment diagram of an element with shear reinforcement based on art. 9.2.1.3 (2) [mm].
            Use your own implementation of this value or use the Form9Dot2ShiftInMomentDiagram class.
        z: MM
            [z] The internal lever arm for an element with constant height, corresponding to the bending moment in the considered element. In the
            shear force calculation of reinforced concrete without axial force, the approximate value z = 0.9d may generally be used [mm].
        n_ed: KN
            [NEd] Design value of axial force [kN].
        """
        super().__init__()
        self.v_ed = v_ed
        self.a_l = a_l
        self.z = z
        self.n_ed = n_ed

    @staticmethod
    def _evaluate(
        v_ed: KN,
        a_l: MM,
        z: MM,
        n_ed: KN,
    ) -> KN:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(z=z, a_l=a_l)
        return abs(v_ed) * a_l / z + n_ed



"""Formula 9.5N from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailing of members and particular rules."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA
from blueprints.validations import raise_if_negative


class Form9Dot5nMinimumShearReinforcementRatio(Formula):
    """Class representing the formula 9.5N for the calculation of the minimum shear reinforcement ratio for beams."""

    label = "9.5N"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_ck: MPA,
        f_yk: MPA,
    ) -> None:
        """[ρw,min] Minimum shear reinforcement ratio for beams [-].

        NEN-EN 1992-1-1+C2:2011 art.9.2.2(5) - Formula (9.5N)

        Parameters
        ----------
        f_ck: MPA
            [fck] Characteristic concrete compressive cylinder strength at 28 days [MPa].
        f_yk: MPA
            [fyk] Characteristic yield strength reinforcement steel [MPa].
        """
        super().__init__()
        self.f_ck = f_ck
        self.f_yk = f_yk

    @staticmethod
    def _evaluate(f_ck: MPA, f_yk: MPA) -> float:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(f_ck=f_ck, f_yk=f_yk)
        return (0.08 * np.sqrt(f_ck)) / f_yk



"""Formula 3.18 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA


class Form3Dot18CompressiveStressConcrete(Formula):
    """Class representing formula 3.18 for the calculation of compressive stress in concrete using stress-strain diagram of figure 3.3."""

    label = "3.18"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_cd: MPA,
    ) -> None:
        """[σc] Compressive stress in concrete using stress-strain diagram of figure 3.3 [MPa].

        NEN-EN 1992-1-1+C2:2011 art.3.1.7(1) - Formula (3.18)

        Parameters
        ----------
        f_cd : MPA
            [fcd] Design value compressive strength concrete [MPa].

        Returns
        -------
        None
        """
        super().__init__()
        self.f_cd = f_cd

    @staticmethod
    def _evaluate(
        f_cd: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_cd < 0:
            raise ValueError(f"Invalid f_cd: {f_cd}. f_cd cannot be negative")
        return f_cd



"""Formula 5.3b from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN
from blueprints.validations import raise_if_negative


class Form5Dot3bTransverseForceBracedMembers(Formula):
    """Class representing formula 5.3b for the calculation of the transverse force for braced members, :math:`H_{i}`.

    See Figure 5.1 a2.
    """

    label = "5.3b"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        theta_i: DIMENSIONLESS,
        n_axial_force: KN,
    ) -> None:
        """[:math:`H_{i}`] Transverse force for braced members [:math:`kN`].

        NEN-EN 1992-1-1+C2:2011 art.5.2(7) - Formula (5.3b)

        Parameters
        ----------
        theta_i : DIMENSIONLESS
            [:math:`Θ_{i}`] Eccentricity, initial inclination imperfections [-].

            Use your own implementation of this value or use the :class:`Form5Dot1Imperfections` class.
        n_axial_force : KN
            [:math:`N`] Axial force [:math:`kN`].

            Positive values for compression, tension is not allowed.

        Notes
        -----
        Eccentricity is suitable for statically determinate members, whereas transverse load can be used for
        both determinate and indeterminate members. The force :math:`H_{i}` may be substituted by some other equivalent
        transverse action.
        """
        super().__init__()
        self.theta_i = theta_i
        self.n_axial_force = n_axial_force

    @staticmethod
    def _evaluate(
        theta_i: DIMENSIONLESS,
        n_axial_force: KN,
    ) -> KN:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(theta_i=theta_i, n_axial_force=n_axial_force)
        return 2 * theta_i * n_axial_force

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.3b."""
        return LatexFormula(
            return_symbol=r"H_{i}",
            result=f"{self:.3f}",
            equation=r"2\theta_{i}N",
            numeric_equation=rf"2\cdot{self.theta_i:.3f}\cdot{self.n_axial_force:.2f}",
            comparison_operator_label="=",
        )



"""Formula 3.16 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.type_alias import MPA


class Form3Dot16DesignValueTensileStrength(Formula):
    """Class representing formula 3.16 for the calculation of the concrete tensile strength design value."""

    label = "3.16"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        alpha_ct: float,
        f_ctk_0_05: MPA,
        gamma_c: float,
    ) -> None:
        """[fcd] Design value concrete tensile strength [MPa].

        NEN-EN 1992-1-1+C2:2011 art.3.1.6(2) - Formula (3.16)

        Parameters
        ----------
        alpha_ct : float
            [αct] Coefficient taking long term effects on tensile strength into account and unfavorable effect due to positioning loading [-]
            See national appendix. Recommended value: 1.0
        f_ctk_0_05 : MPA
            [fctk,0,05] Characteristic tensile strength 5% [MPa].
        gamma_c : float
            [γc] Partial safety factor concrete, see 2.4.2.4 [-].

        Returns
        -------
        None
        """
        super().__init__()
        self.alpha_ct = alpha_ct
        self.f_ctk_0_05 = f_ctk_0_05
        self.gamma_c = gamma_c

    @staticmethod
    def _evaluate(
        alpha_ct: float,
        f_ctk_0_05: MPA,
        gamma_c: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if alpha_ct < 0:
            raise ValueError(f"Invalid alpha_ct: {alpha_ct}. alpha_ct cannot be negative")
        if f_ctk_0_05 < 0:
            raise ValueError(f"Invalid f_ctk_0_05: {f_ctk_0_05}. f_ctk_0_05 cannot be negative")
        if gamma_c <= 0:
            raise ValueError(f"Invalid gamma_c: {gamma_c}. gamma_c cannot be negative or zero")
        return alpha_ct * f_ctk_0_05 / gamma_c



"""Module to keep track of type aliases used in Blueprints."""

# <editor-fold desc="DISTANCES">
MM = float
CM = float
M = float
KM = float
# </editor-fold>

# <editor-fold desc="ROTATIONS">
DEG = float
RAD = float
# </editor-fold>

# <editor-fold desc="WEIGHTS">
KG = float
G = float
# </editor-fold>

# <editor-fold desc="FORCES">
N = float
KN = float
KN_M = float
# </editor-fold>

# <editor-fold desc="MOMENTS">
NM = float
KNM = float
# </editor-fold>

# <editor-fold desc="STRESSES">
MPA = float
KPA = float
GPA = float
# </editor-fold>

# <editor-fold desc="AREAS">
MM2 = float
CM2 = float
DM2 = float
M2 = float
KM2 = float
# </editor-fold>

# <editor-fold desc="VOLUMES">
MM3 = float
CM3 = float
M3 = float
# </editor-fold>

# <editor-fold desc="BI-QUADRATIC">
MM4 = float
CM4 = float
M4 = float
# </editor-fold>

# <editor-fold desc="TIME">
HOURS = float
MINUTES = float
SECONDS = float
DAYS = float
WEEKS = float
MONTHS = float
YEARS = float
# </editor-fold>

# <editor-fold desc="Ratio's">
RATIO = float
PERCENTAGE = float
DIMENSIONLESS = float
# </editor-fold>
